# %%
# import keras
import ast
import json
import os
import re
import shutil
from pathlib import Path

from framework_models.ml_function_classifier import MLFunctionClassifier

ml_function_classifier = MLFunctionClassifier.MLFunctionClassifier()

script_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))

DATA_SCIECE_LIBRARY_DIR = f"{script_dir}/.env/lib/python3.10/site-packages"

LIBRARY_FUNTIONS = {}
out_path = f"{script_dir}/framework_models/_model"

try:
    shutil.rmtree(out_path)
except OSError:
    print("Removal of the directory %s failed" % out_path)

try:
    os.mkdir(out_path)
except OSError:
    print("Creation of the directory %s failed" % out_path)
else:
    print("Successfully created the directory %s " % out_path)


ML_MODULES = {
    "pandas": os.path.join(DATA_SCIECE_LIBRARY_DIR, "pandas"),
    "numpy": os.path.join(DATA_SCIECE_LIBRARY_DIR, "numpy"),
    "matplotlib": os.path.join(DATA_SCIECE_LIBRARY_DIR, "matplotlib"),
    "keras": os.path.join(DATA_SCIECE_LIBRARY_DIR, "keras"),
    "sklearn": os.path.join(DATA_SCIECE_LIBRARY_DIR, "sklearn"),
    "tensorflow": os.path.join(DATA_SCIECE_LIBRARY_DIR, "tensorflow"),
    "lightgbm": os.path.join(DATA_SCIECE_LIBRARY_DIR, "lightgbm"),
    "plotly": os.path.join(DATA_SCIECE_LIBRARY_DIR, "plotly"),
    "scipy": os.path.join(DATA_SCIECE_LIBRARY_DIR, "scipy"),
    "xgboost": os.path.join(DATA_SCIECE_LIBRARY_DIR, "xgboost"),
    "statsmodels": os.path.join(DATA_SCIECE_LIBRARY_DIR, "statsmodels"),
}

LIBRARY_FUNTIONS = {k: {} for k in ML_MODULES}


def get_relative_module_name(filepath, module_name, module_path):
    module_string = []
    for i in range(1, len(filepath.parts) + 1):
        if filepath.parts[-i].startswith("__init__"):
            if str(filepath.parent) == module_path:
                module_string.append(module_name)
                break
        elif filepath.parts[-i] == Path(module_path).stem:
            # module_string.append(module_name)
            break
        else:
            if filepath.parts[-i].endswith(".py"):
                module_string.append(filepath.stem)
            else:
                module_string.append(filepath.parts[-i])

    return ".".join(reversed(module_string))


for _library, _module_path in ML_MODULES.items():
    print(f"Processing: {_library}")
    for _file in Path(_module_path).rglob("*.py"):
        module_name = get_relative_module_name(_file, _library, _module_path)
        source = open(_file).read()

        functions = [
            f"{_library}.{module_name}.{f.name}"
            for f in ast.parse(source).body
            if isinstance(f, ast.FunctionDef)
        ]

        functions_def = {
            f"{_library}.{module_name}.{f.name}": ast.get_docstring(f)
            for f in ast.parse(source).body
            if isinstance(f, ast.FunctionDef)
        }

        classes = [f for f in ast.parse(source).body if isinstance(f, ast.ClassDef)]

        for _class in classes:
            functions.extend(
                [
                    f"{_library}.{module_name}.{_class.name}.{f.name}"
                    for f in _class.body
                    if isinstance(f, ast.FunctionDef)
                ]
            )

            class_function_defs = {
                f"{_library}.{module_name}.{_class.name}.{f.name}": ast.get_docstring(f)
                for f in _class.body
                if isinstance(f, ast.FunctionDef)
            }

        for _class in classes:
            functions.extend(
                [f"{_library}.{module_name}.{_class.name}" for f in _class.body]
            )

            class_function_defs = {
                f"{_library}.{module_name}.{_class.name}": ast.get_docstring(f)
                for f in _class.body
                if isinstance(f, ast.ClassDef)
            }

            functions_def = functions_def | class_function_defs

        if module_name == "io.parsers.readers":
            print()

        LIBRARY_FUNTIONS[_library][module_name] = {
            k: ml_function_classifier.predict_function(k, str(d))
            for k, d in functions_def.items()
        }
        json_path = f'{out_path}/{_library}/{module_name.replace(".","/")}.json'
        try:
            os.makedirs(re.sub("(\/[^\/]+)\/?$", "", json_path))
        except FileExistsError:
            pass
            # print("Directory already exists")

        with open(f"{json_path}", "w") as outfile:
            json.dump(LIBRARY_FUNTIONS[_library][module_name], outfile, indent=4)

print()



import json
import os
import pathlib
import pickle
from pathlib import Path

import numpy
import pygtrie

from framework_models.ml_function_classifier import MLFunctionClassifier

ml_function_classifier = MLFunctionClassifier.MLFunctionClassifier()
ML_PIPELINE_MODEL = True
CACHE_PREDICTIONS = {}

SCRIPT_ROOT = pathlib.Path(__file__).parent.absolute()
cache_models = True
# cache_models = True

MODELS = {}
PHASES = {
    "IMPORTS": "Imports",
    "CONFIGURATION": "Configuration",
    "FUNCTION_DEFINITION": "Function Definition",
    "DECLARED_FUNCTION": "Declared Function",
    "BUILTIN_FUNCTION": "Builtin Function",
    "LIBRARY_LOADING": "Library Loading",
    "DATA_LOADING": "Data Loading",
    "VISUALIZATION": "Visualization",
    "UNKNOWN": "Unknown",
    "DATA_CLEANING_PREPARATION": "Data Preparation",
    "DATA_PROFILING_AND_EXPLORATORY_DATA_ANALYSIS": "Data Profiling and Exploratory Data Analysis",
    "DATA_CLEANING_FILTERING": "Data Cleaning Filtering",
    "DATA_SUB_SAMPLING_AND_TRAIN_TEST_SPLITTING": "Data Sub-sampling and Train-test Splitting",
    "FEATURE_ENGINEERING": "Feature Engineering",
    "FEATURE_TRANSFORMATION": "Feature Transformation",
    "FEATURE_SELECTION": "Feature Selection",
    "MODEL_BUILDING_AND_TRAINING": "Model Building and Training",
    "MODEL_TRAINING": "Model Training",
    "MODEL_PARAMETER_TUNING": "Model Parameter Tuning",
    "MODEL_VALIDATION_AND_ASSEMBLING": "Model Validation and Assembling",
}

PHASES_EXPLANATION = {
    # "IMPORTS": "Imports",
    # "CONFIGURATION": "Configuration",
    # "FUNCTION_DEFINITION": "Function Definition",
    # "DECLARED_FUNCTION": "Declared Function",
    # "BUILTIN_FUNCTION": "Builtin Function",
    # "LIBRARY_LOADING": "Library Loading",
    "DATA_LOADING": "The process of acquiring data that will be used to build a model",
    "VISUALIZATION": "Visualization of data using plots and graphical techniques",
    # "UNKNOWN": "Unknown",
    # "DATA_CLEANING_PREPARATION": "Data Cleaning Preparation",
    "DATA_PROFILING_AND_EXPLORATORY_DATA_ANALYSIS": "Assess overall data quality, detect noise in labels, skewness and correlations, class imbalance, and target distributions, bias detection",
    "DATA_CLEANING_FILTERING": "Auto Cleaning or filtering, filling missing values, removing outliers, bias mitigation",
    "DATA_SUB_SAMPLING_AND_TRAIN_TEST_SPLITTING": "Determine the best strategy for train-test splitting and sub-sampling methodology for large data",
    # "FEATURE_ENGINEERING": "Feature Engineering",
    "FEATURE_TRANSFORMATION": "Determine a ranked list of features that should be added or dropped, determine the best encoding strategy for certain features automatically, identify column concepts and link them to derive useful new features based on domain knowledge ",
    "FEATURE_SELECTION": "Select the best subset of features",
    # "MODEL_BUILDING_AND_TRAINING": "Model Building and Training",
    "MODEL_TRAINING": "The process of applying the right model to fit to the data and train the model",
    "MODEL_PARAMETER_TUNING": "Hypter parameter choosing and tuning",
    "MODEL_VALIDATION_AND_ASSEMBLING": "The process of ensuring the generalizability of a model, evaluating and comparing models, assembling multiple models",
}

PHASE_GROUPS = {
    "Library Loading": ["Library Loading"],
    "Visualization": ["Visualization"],
    "Others": ["Others"],
    "Data Preparation": [
        "Data Preparation",
        "Data Profiling and Exploratory Data Analysis",
        "Data Preparation and Exploration",
        "Data Cleaning Filtering",
        "Data Sub-sampling and Train-test Splitting",
        "Data Loading",
        "Exploratory Data Analysis",
    ],
    "Feature Engineering": [
        "Feature Engineering",
        "Feature Transformation",
        "Feature Selection",
    ],
    "Model Building and Training": [
        "Model Building and Training",
        "Model Training",
        "Model Parameter Tuning",
        "Model Validation and Assembling",
    ],
}


def get_high_level_phase(phase):
    for _k, _v in PHASE_GROUPS.items():
        if phase in _v:
            return _k

    return PHASES["UNKNOWN"]


ML_MODULES = {
    "keras": os.path.join(SCRIPT_ROOT, "model", "keras"),
    "lightgbm": os.path.join(SCRIPT_ROOT, "model", "lightgbm"),
    "matplotlib": os.path.join(SCRIPT_ROOT, "model", "matplotlib"),
    "numpy": os.path.join(SCRIPT_ROOT, "model", "numpy"),
    "pandas": os.path.join(SCRIPT_ROOT, "model", "pandas"),
    "plotly": os.path.join(SCRIPT_ROOT, "model", "plotly"),
    "scipy": os.path.join(SCRIPT_ROOT, "model", "scipy"),
    "seaborn": os.path.join(SCRIPT_ROOT, "model", "seaborn"),
    "sklearn": os.path.join(SCRIPT_ROOT, "model", "sklearn"),
    "xgboost": os.path.join(SCRIPT_ROOT, "model", "xgboost"),
    "statsmodels": os.path.join(SCRIPT_ROOT, "model", "statsmodels"),
    "tensorflow": os.path.join(SCRIPT_ROOT, "model", "tensorflow")
    # "tensorflow": os.path.join(SCRIPT_ROOT, "tensorflow"),
}

ML_MODULES_ALIAS = {
    "keras": json.loads(
        open(os.path.join(SCRIPT_ROOT, "aliases", "keras.json")).read()
    ),
    "matplotlib": json.loads(
        open(os.path.join(SCRIPT_ROOT, "aliases", "matplotlib.json")).read()
    ),
    "numpy": json.loads(
        open(os.path.join(SCRIPT_ROOT, "aliases", "numpy.json")).read()
    ),
    "pandas": json.loads(
        open(os.path.join(SCRIPT_ROOT, "aliases", "pandas.json")).read()
    ),
    "tensorflow": json.loads(
        open(os.path.join(SCRIPT_ROOT, "aliases", "tensorflow.json")).read()
    ),
    # "tensorflow": os.path.SCRIPT_ROOT, "tensorflow"),
}

MODELS = {k: pygtrie.StringTrie(separator=".") for k in ML_MODULES}


def get_relative_module_name(filepath, module_name, module_path):
    module_string = []
    for i in range(1, len(filepath.parts) + 1):
        if filepath.parts[-i].startswith("__init__"):
            if str(filepath.parent) == module_path:
                module_string.append(module_name)
                break
        elif filepath.parts[-i] == Path(module_path).stem:
            module_string.append(module_name)
            break
        else:
            if filepath.parts[-i].endswith(".json"):
                module_string.append(filepath.stem)
            else:
                module_string.append(filepath.parts[-i])

    return ".".join(reversed(module_string))


if cache_models and os.path.exists(os.path.join(SCRIPT_ROOT, "models_cache.pickle")):
    MODELS = pickle.load(open(os.path.join(SCRIPT_ROOT, "models_cache.pickle"), "rb"))
else:
    for _module_name, _module_path in ML_MODULES.items():
        counter = 0
        for _file in Path(_module_path).rglob("*.json"):
            # if "numpy" in _file.name:
            #     print()
            print(counter, _file)
            module_name = get_relative_module_name(_file, _module_name, _module_path)
            MODELS[_module_name][module_name] = json.loads(open(_file).read())
            counter += 1

    if cache_models:
        pickle.dump(
            MODELS, open(os.path.join(SCRIPT_ROOT, "models_cache.pickle"), "wb")
        )


def check_alias(func_call):
    root_module = func_call.split(".")[0]
    if root_module not in MODELS:
        return func_call

    if root_module in ML_MODULES_ALIAS:
        if func_call in ML_MODULES_ALIAS[root_module]:
            func_call = ML_MODULES_ALIAS[root_module][func_call]

    return func_call


def lookup_pipeline_tag(func_call, doc_string=""):
    if ML_PIPELINE_MODEL:
        # Check if its available from cache, else try ml
        if func_call in CACHE_PREDICTIONS:
            return CACHE_PREDICTIONS[func_call]
        else:
            _res = lookup_pipeline_tag_ml(func_call, doc_string)
            CACHE_PREDICTIONS[func_call] = _res
            return _res

    else:
        return lookup_pipeline_tag_hg(func_call)


def lookup_pipeline_tag_hg(func_call):
    root_module = func_call.split(".")[0]
    if root_module not in MODELS:
        return [PHASES["UNKNOWN"]]

    if root_module in ML_MODULES_ALIAS:
        if func_call in ML_MODULES_ALIAS[root_module]:
            func_call = ML_MODULES_ALIAS[root_module][func_call]

    try:
        model_match = MODELS[root_module].longest_prefix(func_call)
        _res = [PHASES[k] for k in model_match.value[func_call]]
        if _res:
            return _res
        else:
            return [PHASES["UNKNOWN"]]
    except:
        # print("ML tag missing for:", func_call)
        return [PHASES["UNKNOWN"]]


def lookup_pipeline_tag_ml(func_call, doc_string):
    root_module = func_call.split(".")[0]

    if root_module in ML_MODULES_ALIAS:
        if func_call in ML_MODULES_ALIAS[root_module]:
            func_call = ML_MODULES_ALIAS[root_module][func_call]

    try:
        _res = ml_function_classifier.predict_function(func_call, doc_string)
        # _classic_res = lookup_pipeline_tag(func_call)

        if not _res:
            _res = [PHASES["UNKNOWN"]]
    except:
        print("ML tag missing for:", func_call)
        return [PHASES["UNKNOWN"]]

    return _res


if __name__ == "__main__":
    for line in open("test.txt").readlines():
        # print(line.strip())
        # print(lookup_pipeline_tag(line.strip()))

        lookup_pipeline_tag(line.strip())
        # lookup_pipeline_tag("keras.backend.get_session")


"""
Pipeline Phases

LIBRARY_LOADING
DATA_LOADING
VISUALIZATION
UNKNOWN
DATA_CLEANING_PREPARATION
DATA_PROFILING_AND_EXPLORATORY_DATA_ANALYSIS
DATA_CLEANING_FILTERING
DATA_SUB_SAMPLING_AND_TRAIN_TEST_SPLITTING
FEATURE_ENGINEERING
FEATURE_TRANSFORMATION
FEATURE_SELECTION
MODEL_BUILDING_AND_TRAINING
MODEL_TRAINING
MODEL_PARAMETER_TUNING
MODEL_VALIDATION_AND_ASSEMBLING
"""

# MODEL_SAVING_AND_LOADING



import csv
import json
import os
import subprocess
import sys
from pathlib import Path

from pycg_extended import pycg


def do_sorted(d):
    s = {}
    for n in d:
        s[str(n)] = sorted(d[n])
    return s


def get_subdirs(item):
    return os.listdir(item)


def equal_sound(out, expected):
    for item in expected:
        if not item in out:
            return False
        for edge in expected[item]:
            if not edge in out[item]:
                return False

    return True


def equal_complete(out, expected):
    for item in out:
        if not item in expected:
            continue
        for edge in out[item]:
            if not edge in expected[item]:
                return False

    return True


def get_python_files(test):
    return sorted(Path(test).rglob("*.py"))


def write_csv(res_file, data):
    header = ["Category", "Num-Cases", "Complete", "Sound"]
    with open(res_file, "w+") as f:
        writer = csv.writer(f, delimiter=",")
        writer.writerow(header)
        for cat in sorted(data.keys()):
            writer.writerow(
                [cat, data[cat]["all"], data[cat]["complete"], data[cat]["sound"]]
            )


def iterate_cats(test_suite_dir, ex, results_file, do_test):
    overall_complete = 0
    overall_sound = 0
    all_tests = 0
    data = {}
    for cat in sorted(get_subdirs(test_suite_dir)):
        if cat in ["context_sensitive", "dynamic", "external", "task_test"]:
            continue

        # # Debug
        # ['args', 'assignments', 'builtins', 'classes', 'context_sensitive', 'decorators', 'dicts', 'direct_calls', 'dynamic', 'exceptions', 'external', 'flow_sensitive', 'functions', 'generators', 'imports', 'kwargs', 'lambdas', 'lists', 'mro', 'returns', 'task_test']
        # if cat not in ["returns"]:
        #     continue

        print("Iterating category {}...".format(cat))
        complete_passed = 0
        sound_passed = 0
        tests = get_subdirs(os.path.join(test_suite_dir, cat))
        for test in tests:
            print(f"Running {test}...")
            test_path = os.path.join(test_suite_dir, cat, test)

            _result_actual, _result_expected = do_test(test_path)

            if equal_complete(_result_actual, _result_expected):
                complete_passed += 1
            else:
                # print(f"##- Not Complete {test}")
                pass

            if equal_sound(_result_actual, _result_expected):
                sound_passed += 1
            else:
                pass

        data[cat] = {
            "complete": complete_passed,
            "sound": sound_passed,
            "all": len(tests),
        }
        overall_complete += complete_passed
        overall_sound += sound_passed
        all_tests += len(tests)
    write_csv(results_file, data)


def do_test_hg(test):
    cg_path = os.path.join(test, "linesCallSite.json")
    main_path = os.path.join(test, "main.py")

    cs = pycg.CallGraphGenerator([main_path], test, -1, "call-site")
    cs.analyze()

    out_cg = cs.output_call_sites(module_name="main")

    with open(cg_path) as f:
        expected_cg = json.loads(f.read())

    return do_sorted(out_cg), do_sorted(expected_cg)


def main():
    test_suite_dir = "/app/HeaderGen/callsites-jupyternb-micro-benchmark/snippets"
    results_dir = "/results"

    hg_results = os.path.join(results_dir, "headergen_micro_benchmark_eval.csv")

    print("-" * 40)
    print("Iterating categories for Headergen")
    print("-" * 40)
    iterate_cats(test_suite_dir, None, hg_results, do_test_hg)
    print("\n")


if __name__ == "__main__":
    main()



import os
import pickle

import numpy as np

from framework_models.ml_function_classifier.data_cleaning import data_cleaning

script_dir = os.path.abspath(os.path.dirname(__file__))


class MLFunctionClassifier:
    def __init__(self):
        with open(f"{script_dir}/highlevel_vectorizer.pkl", "rb") as f:
            self.highlevel_vectorizer = pickle.load(f)

        with open(f"{script_dir}/highlevel_classifier.pkl", "rb") as f:
            self.highlevel_classifier = pickle.load(f)

        with open(f"{script_dir}/dp_vectorizer.pkl", "rb") as f:
            self.dp_vectorizer = pickle.load(f)

        with open(f"{script_dir}/dp_classifier.pkl", "rb") as f:
            self.dp_classifier = pickle.load(f)

        with open(f"{script_dir}/fe_vectorizer.pkl", "rb") as f:
            self.fe_vectorizer = pickle.load(f)

        with open(f"{script_dir}/fe_classifier.pkl", "rb") as f:
            self.fe_classifier = pickle.load(f)

        with open(f"{script_dir}/mb_vectorizer.pkl", "rb") as f:
            self.mb_vectorizer = pickle.load(f)

        with open(f"{script_dir}/mb_classifier.pkl", "rb") as f:
            self.mb_classifier = pickle.load(f)

        self.dp_labels = [
            "Data Cleaning Filtering",
            "Data Loading",
            "Data Preparation and Exploration",
            "Data Sub-sampling and Train-test Splitting",
            "Exploratory Data Analysis",
        ]

        self.fe_labels = [
            "Feature Engineering",
            "Feature Selection",
            "Feature Transformation",
        ]

        self.mb_labels = [
            "Model Building and Training",
            "Model Parameter Tuning",
            "Model Training",
            "Model Validation and Assembling",
        ]

    def predict_dp(self, preprocessed_docstring):
        dp_numerical_vector = self.dp_vectorizer.transform([preprocessed_docstring])
        dp_predicted_probabilities = self.dp_classifier.predict_proba(
            dp_numerical_vector
        )
        dp_predicted_label = (dp_predicted_probabilities > 0.5).astype(int)
        dp_indices = np.where(np.array(dp_predicted_label) == 1)[1]
        dp_predicted_classes = [self.dp_labels[i] for i in dp_indices]
        return dp_predicted_classes

    def predict_fe(self, preprocessed_docstring):
        fe_numerical_vector = self.fe_vectorizer.transform([preprocessed_docstring])
        fe_predicted_probabilities = self.fe_classifier.predict_proba(
            fe_numerical_vector
        )
        fe_predicted_label = (fe_predicted_probabilities > 0.5).astype(int)
        fe_indices = np.where(np.array(fe_predicted_label) == 1)[1]
        fe_predicted_classes = [self.fe_labels[i] for i in fe_indices]
        return fe_predicted_classes

    def predict_mb(self, preprocessed_docstring):
        mb_numerical_vector = self.mb_vectorizer.transform([preprocessed_docstring])
        mb_predicted_probabilities = self.mb_classifier.predict_proba(
            mb_numerical_vector
        )
        mb_predicted_label = (mb_predicted_probabilities > 0.5).astype(int)
        mb_indices = np.where(np.array(mb_predicted_label) == 1)[1]
        mb_predicted_classes = [self.mb_labels[i] for i in mb_indices]
        return mb_predicted_classes

    def predict_function(self, func_call, docstring):
        root = func_call.split(".")[0]
        if root in ["warnings", "os", "posix"]:
            return []

        preprocessed_docstring = data_cleaning(docstring)
        if len(preprocessed_docstring.split()) < 5:
            return []

        highlevel_numerical_vector = self.highlevel_vectorizer.transform(
            [preprocessed_docstring]
        )
        highlevel_predicted_probabilities = self.highlevel_classifier.predict_proba(
            highlevel_numerical_vector
        )
        highlevel_predicted_label = (highlevel_predicted_probabilities > 0.5).astype(
            int
        )
        if (
            highlevel_predicted_probabilities[0][0] > 0.5
            and highlevel_predicted_probabilities[0][1] > 0.3
        ):
            highlevel_predicted_label[0][1] = 1

        predicted_classes = []

        if highlevel_predicted_label[0][0] == 1:
            dp_classes = self.predict_dp(preprocessed_docstring)

            if len(dp_classes) == 1:
                predicted_classes.extend(dp_classes)
            else:
                predicted_classes.append("Data Preparation and Exploration")

        if highlevel_predicted_label[0][1] == 1:
            fe_classes = self.predict_fe(preprocessed_docstring)
            if len(fe_classes) == 1:
                predicted_classes.extend(fe_classes)
            else:
                predicted_classes.append("Feature Engineering")

        if highlevel_predicted_label[0][2] == 1:
            mb_classes = self.predict_mb(preprocessed_docstring)
            if len(mb_classes) == 1:
                predicted_classes.extend(mb_classes)
            else:
                predicted_classes.append("Model Building and Training")

        if highlevel_predicted_label[0][3] == 1:
            predicted_classes.append("Visualization")

        return predicted_classes




import os

from base import TestBase

class MachineLearningTest(TestBase):
    snippet_dir = "machine_learning"

    def test_return_type(self):
        self.validate_snippet(self.get_snippet_path("return_type"))




import os

from base import TestBase

class TaskTestTest(TestBase):
    snippet_dir = "task_test"

    def test_custom(self):
        self.validate_snippet(self.get_snippet_path("custom"))

    def test_expanded(self):
        self.validate_snippet(self.get_snippet_path("expanded"))




import os

from base import TestBase

class ContextSensitiveTest(TestBase):
    snippet_dir = "context_sensitive"

    def test_attribute(self):
        self.validate_snippet(self.get_snippet_path("attribute"))



import jupytext
import re
from pathlib import Path


def create_input_py(filename):
    py_ntbk_path = None
    if filename.endswith(".ipynb"):
        _file = Path(filename) 
        _filename = _file.name.split(".ipynb")[0]
        ntbk = jupytext.read(_file)
        # py_ntbk = jupytext.writes(ntbk, fmt='py:percent')
        py_ntbk_path = "{}/{}.py".format(Path(_file).parent, _filename)

        # write to python file for analysis
        jupytext.write(ntbk, py_ntbk_path, fmt='py:percent')
        filename = py_ntbk_path

    return filename


# Find all blocks and the line numbers in notebook script
def find_block_numbers(filename):
    if filename.endswith(".ipynb"):
        _file = Path(filename) 
        _filename = _file.name.split(".ipynb")[0]
        ntbk = jupytext.read(_file)
        py_ntbk = jupytext.writes(ntbk, fmt='py:percent')
        py_source_split = py_ntbk.split("\n")
    else:
        py_source_split = filename.split("\n")

    _start, _end = None, None
    lineno = 1
    block = 1
    mapping = {}

    _current_md = False
    for _line in py_source_split:
        if _line.startswith("# %%"):
            if _start is None:
                _start = lineno
                if _line.startswith("# %% [markdown]"):
                    _current_md = True
                else:
                    _current_md = False
            else:
                _end = lineno
                if _end == (_start+1):
                    _start = lineno
                    continue

                if not _current_md:
                    mapping[block] = {
                        "start": _start,
                        "end": _end - 1
                    }
                    block += 1

                _start = _end
                if _line.startswith("# %% [markdown]"):
                    _current_md = True
                else:
                    _current_md = False

        lineno += 1

    if not _current_md:    
        mapping[block] = {
            "start": _start,
            "end": lineno - 1
        }
    return mapping


def get_block_of_lineno(lineno, block_mapping):
    for map_key, map_value in block_mapping.items():
        if map_value["start"] <= lineno <= map_value["end"]:
            return map_key

    return None

def get_cell_numbers(call_sites, filename, module_name):
    block_mapping = find_block_numbers(filename)
    cell_call_sites = {}
    def cellid_repl(matchobj):
        _cell_id = get_block_of_lineno(int(matchobj.group(0)), block_mapping)
        return str(_cell_id)

    for _cs_line, _cs_calls in call_sites.items():
        # (:[\s\S]*?\.)|:[\s\S]*?$
        # :(.*)\.|:(.*)$
        _cell_id = get_block_of_lineno(int(_cs_line), block_mapping)
        if _cell_id not in cell_call_sites:
            cell_call_sites[_cell_id] = set()

        # only keep cellid of calls found in the same notebook
        cell_calls = [re.sub(r"(?s):.*?(?=\.)|:.*?(?=$)", "", _call) if not _call.startswith(module_name) 
                            else re.sub(r"(?s)(?<=:).*?(?=\.)|(?<=:).*?(?=$)", cellid_repl, _call) for _call in _cs_calls]

        cell_call_sites[_cell_id] = cell_call_sites[_cell_id].union(cell_calls)

    return cell_call_sites


#
# Copyright (c) 2020 Vitalis Salis.
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import sys
import importlib
import json
import utils

from unittest import TestCase, main


class TestBase(TestCase):
    snippet_dir = ""

    def setUp(self):
        def error():
            print("Invalid module %s.%s" % (cg_mod, cg_class))
            print(
                "Set environment variables `CALL_GRAPH_CLASS` and `CALL_GRAPH_MODULE` properly"
            )
            sys.exit(1)

        self.snippets_path = os.environ.get("SNIPPETS_PATH")
        self.notebooks_path = os.environ.get("NOTEBOOKS_PATH")
        cg_class = "CallGraphGenerator"
        cg_mod = "pycg_extended.pycg"
        if not cg_class or not cg_mod:
            error()
        try:
            self.cg_mod = importlib.import_module(cg_mod)
        except ImportError:
            error()

        self.cg_class = getattr(self.cg_mod, cg_class)
        if not self.cg_class:
            error()

    def validate_snippet(self, snippet_path):
        output = self.get_snippet_output_cs(snippet_path)
        expected = self.get_snippet_expected_cs(snippet_path)

        def do_sorted(d):
            s = {}
            for n in d:
                s[str(n)] = sorted(d[n])
            return s

        self.assertEqual(output, expected)

    def get_snippet_path(self, name):
        return os.path.join(self.snippets_path, self.snippet_dir, name)

    def get_snippet_output_cg(self, snippet_path):
        main_path = os.path.join(snippet_path, "main.py")
        try:
            cg = self.cg_class([main_path], snippet_path, -1, "call-graph")
            cg.analyze()
            return cg.output()
        except Exception as e:
            cg.tearDown()
            raise e

    def get_snippet_output_cs(self, snippet_path):
        main_path = os.path.join(snippet_path, "main.py")
        try:
            cs = self.cg_class([main_path], snippet_path, -1, "call-site")
            cs.analyze()
            # graphs = self.get_json_graph(cs)
            return cs.output_call_sites(module_name="main")
        except Exception as e:
            cs.tearDown()
            raise e

    def validate_notebook(self, snippet_path):
        output = self.get_notebook_output_cs(snippet_path)
        expected = self.get_notebook_expected_cs(snippet_path)

        self.assertEqual(output, expected)

    def get_notebook_expected_cs(self, snippet_path):
        cg_path = os.path.join(snippet_path, "cellsCallSite.json")
        with open(cg_path, "r") as f:
            return json.loads(f.read())

    def get_notebook_output_cs(self, snippet_path):
        main_nb_path = os.path.join(snippet_path, "main_nb.ipynb")
        # TODO: notebook conversion should be part of pycg
        main_path = utils.create_input_py(main_nb_path)
        try:
            cs = self.cg_class([main_path], snippet_path, -1, "call-site")
            cs.analyze()
            output_callsites = cs.output_call_sites(module_name="main_nb")

            return utils.get_cell_numbers(
                output_callsites, main_nb_path, module_name="main_nb"
            )
        except Exception as e:
            cs.tearDown()
            raise e
        finally:
            os.remove(main_path)

    def get_json_graph(self, cg):
        from pycg_extended import formats
        from collections import OrderedDict

        formatter = formats.Simple(cg)
        as_formatter = formats.AsGraph(cg)

        fa_formatter = formats.Fasten(cg, None, "", "", "", 0)

        cs_formatter = formats.CallSites(cg)

        cg_json = formatter.generate()
        ag_json = as_formatter.generate()
        fa_json = fa_formatter.generate()

        cs_json = cs_formatter.generate()
        cs_json = cs_formatter.get_cg()
        # cs_json = cs_formatter.get_ag()

        return (
            OrderedDict(sorted(cg_json.items())),
            OrderedDict(sorted(ag_json.items())),
        )

    def get_snippet_expected_cg(self, snippet_path):
        cg_path = os.path.join(snippet_path, "callgraph.json")
        with open(cg_path, "r") as f:
            return json.loads(f.read())

    def get_snippet_expected_cs(self, snippet_path):
        cg_path = os.path.join(snippet_path, "linesCallSite.json")
        with open(cg_path, "r") as f:
            return json.loads(f.read())

    def assertEqual(self, actual, expected):
        def do_sorted(d):
            s = {}
            for n in d:
                s[str(n)] = sorted(d[n])
            return s

        super().assertEqual(do_sorted(actual), do_sorted(expected))


if __name__ == "__main__":
    main()



import re

import nltk

nltk.download("stopwords", quiet=True)
nltk.download("punkt", quiet=True)
nltk.download("wordnet", quiet=True)

from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import word_tokenize

stop_words = set(stopwords.words("english"))


def clean_dataframe_text(text):
    # remove pd.DataFrame()
    pattern = r"df\s*=\s*pd\.DataFrame\(((?:\([^()]*\)|[^()])*)\)"
    cleaned_text = re.sub(pattern, "", text, flags=re.DOTALL)

    pattern = r"^\s*df\s*=\s*pd\.DataFrame\({.*},?\s*index=\[.*\]\)\s*$"
    cleaned_text = re.sub(pattern, "", cleaned_text, flags=re.DOTALL)

    # remove >>df and its next line
    pattern = r"^>>> df\s*$\n^.*$\n"
    cleaned_text = re.sub(pattern, "", cleaned_text, flags=re.MULTILINE)

    # 0  Yum Yum   cup     4.0
    pattern = r"^\d+\s+.*$[\n\r]*?"
    cleaned_text = re.sub(pattern, "", cleaned_text, flags=re.MULTILINE)

    # Portland    17.0    62.6

    pattern = r"\n([A-Z][a-z]+)\s+\d+\.\d+(\s+\d+\.\d+){0,2}"
    cleaned_text = re.sub(pattern, "", cleaned_text)

    return cleaned_text.strip()


def remove_latex(text):
    # Remove code blocks starting with `.. math::`
    text = re.sub(
        r"^\s*\.\.\s*math::\n\s*(.*?)\n(\n|$)",
        " ",
        text,
        flags=re.MULTILINE | re.DOTALL,
    )

    # remove include statements

    text = re.sub(r"^\s*..\s+include::\s+.+", " ", text, flags=re.MULTILINE)

    # remove the version added statements such as .. versionadded:: 0.19

    text = re.sub(r"^\s*\.\. versionadded:: .*$", "", text, flags=re.MULTILINE)

    # remove the deprecated statements such as .. deprecated::

    text = re.sub(r"^\s*\.\. deprecated:: .*$", "", text, flags=re.MULTILINE)

    return text


def remove_math_words(text):
    # remove words that starts with :math:`sometext`
    return re.sub(r":math:`[^`]*`", "", text)


def remove_urls(text):
    # remove the urls starts with https:
    return re.sub(r"http\S+", " ", text)


def remove_references(text):
    # References ...
    text = re.sub(r"References\s+-+\s+[\s\S]*", "", text)

    #:ref:``
    text = re.sub(r":ref:`[^`]*`", "", text)

    return text


def remove_single_letter_underscore(text):
    # remove the underscore letters with single letter on each sides such as  h_0, c_0
    pattern = r"\b\w{0,1}_\w{0,1}\b"
    return re.sub(pattern, " ", text)


def remove_non_alpha(text):
    # remove non alpha characters
    cleaned_text = re.sub(r"[^a-zA-Z\s]", " ", text)
    return cleaned_text


def remove_single_letters(text):
    # Remove single letters such as h ,k i
    return re.sub(r"(?<!\S)\w(?!\S)", " ", text)


def remove_stopwords_and_lemmatize(text):
    # remove stop words and lemmatize the text
    word_tokens = word_tokenize(text)
    lemmatizer = WordNetLemmatizer()
    filtered_text = [
        lemmatizer.lemmatize(word)
        for word in word_tokens
        if not word.lower() in stop_words
    ]
    return " ".join(filtered_text).lower()


def data_cleaning(text):
    cleanedtext = remove_latex(text)
    cleanedtext = clean_dataframe_text(cleanedtext)
    cleanedtext = remove_urls(cleanedtext)
    cleanedtext = remove_math_words(cleanedtext)
    cleanedtext = remove_references(cleanedtext)
    cleanedtext = remove_single_letter_underscore(cleanedtext)
    cleanedtext = remove_non_alpha(cleanedtext)
    cleanedtext = remove_single_letters(cleanedtext)
    cleanedtext = remove_stopwords_and_lemmatize(cleanedtext)
    return cleanedtext




import os

from base import TestBase

class FlowSensitiveTest(TestBase):
    snippet_dir = "flow_sensitive"

    def test_conditional(self):
        self.validate_snippet(self.get_snippet_path("conditional"))

    def test_multiple_class(self):
        self.validate_snippet(self.get_snippet_path("multiple_class"))

    def test_multiple_func(self):
        self.validate_snippet(self.get_snippet_path("multiple_func"))

    def test_multi_assign_per_line(self):
        self.validate_snippet(self.get_snippet_path("multi_assign_per_line"))

    def test_reassigned_call(self):
        self.validate_snippet(self.get_snippet_path("reassigned_call"))

    def test_repeated(self):
        self.validate_snippet(self.get_snippet_path("repeated"))

    def test_simple(self):
        self.validate_snippet(self.get_snippet_path("simple"))

    def test_tuples(self):
        self.validate_snippet(self.get_snippet_path("tuples"))
